from rdkit import Chem
from meeko import MoleculePreparation
from meeko import oids_block_from_setup
from meeko import parse_offxml
from meeko import AtomTyper
from meeko import Hydrate
import openforcefields
import pathlib
import json

p = pathlib.Path(openforcefields.__file__) # find openff-forcefields xml files
offxml = p.parents[0] / "offxml" / "openff-2.0.0.offxml"
offxml = offxml.resolve()
print(offxml)
vdw_list, dihedral_list, vdw_by_type = parse_offxml(offxml)
meeko_config = {"keep_nonpolar_hydrogens": True, "flexible_amides": True}
meeko_config["atom_type_smarts"] = json.loads(AtomTyper.defaults_json)
meeko_config["atom_type_smarts"]["ATOM_PARAMS"] = {"openff-2.0.0": vdw_list}
meeko_config["atom_type_smarts"]["DIHEDRALS"] = dihedral_list
mk_prep = MoleculePreparation.from_config(meeko_config)

def mol_from_smiles(smiles):
    mol = Chem.MolFromSmiles(smiles)
    mol = Chem.AddHs(mol)
    etkdg_param = Chem.rdDistGeom.ETKDGv3()
    Chem.rdDistGeom.EmbedMolecule(mol, etkdg_param)
    return mol
    
#def test():
if __name__ == "__main__":
    mol = mol_from_smiles("c1c(O)ccnc1NC(=O)C[NH3+]")
    mk_prep.prepare(mol)
    molsetup = mk_prep.setup
    oids = oids_block_from_setup(molsetup)
    #print(oids)
    hydrate = Hydrate()
    waters = hydrate(molsetup)
    #print(Chem.MolToMolBlock(mol))
    s = molsetup.write_xyz_string().split("\n")[2:-1]
    for w in waters:
        s.extend(w.write_xyz_string().split("\n")[2:-1])
    print("%d\n\n%s" % (len(s), "\n".join(s)), file=open("hydrated.xyz", "w"))




#if __name__ == "__main__":
#    test()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Meeko hydrate molecule
#

import json
import numpy as np
from rdkit import Chem
from rdkit.Geometry import Point3D
import warnings

from .atomtyper import AtomicGeometry
from .molsetup import RDKitMoleculeSetup
from .utils import geomutils
from .utils import pdbutils


class Waters:

    @staticmethod
    def make_molsetup_OPC() -> RDKitMoleculeSetup:
        """
        Builds an RDKitMoleculeSetup for an OPC water.

        Returns
        -------
        molsetup: RDKitMoleculeSetup
            An RDKitMoleculeSetup for an OPC water.
        """

        raise NotImplementedError
        rdmol = Chem.MolFromSmiles("O")
        conformer = Chem.Conformer(rdmol.GetNumAtoms())
        conformer.SetAtomPosition(0, Point3D(0, 0, 0))
        rdmol.AddConformer(conformer)
        molsetup = RDKitMoleculeSetup.from_mol(rdmol)
        return molsetup

    @staticmethod
    def make_molsetup_TIP3P_AA() -> RDKitMoleculeSetup:
        """
        Builds an RDKitMoleculeSetup for an all atom (AA) TIP3P water, i.e. explicit Hs

        Returns
        -------
        molsetup: RDKitMoleculeSetup
            An RDKitMoleculeSetup for an all atom (AA) TIP3P water.
        """

        rdmol = Chem.MolFromSmiles("O")
        rdmol = Chem.AddHs(rdmol)
        conformer = Chem.Conformer(rdmol.GetNumAtoms())
        dist_oh = 0.9572
        ang_hoh = np.radians(104.52)
        conformer.SetAtomPosition(0, Point3D(0, 0, 0))
        conformer.SetAtomPosition(1, Point3D(dist_oh, 0, 0))
        conformer.SetAtomPosition(
            2, Point3D(np.cos(ang_hoh) * dist_oh, np.sin(ang_hoh) * dist_oh, 0)
        )
        rdmol.AddConformer(conformer)
        molsetup = RDKitMoleculeSetup.from_mol(rdmol)
        molsetup.add_bond(0, 1, rotatable=False)
        molsetup.add_bond(0, 2, rotatable=False)
        molsetup.atoms[0].atom_type = "n-tip3p-O"
        molsetup.atoms[1].atom_type = "n-tip3p-H"
        molsetup.atoms[2].atom_type = "n-tip3p-H"
        molsetup.atoms[0].charge = -0.834
        molsetup.atoms[1].charge = 0.417
        molsetup.atoms[2].charge = 0.417
        return molsetup


class Hydrate:
    defaults = [
        {
            "smarts": "[#7X2;v3;!+](=,:[*])[*]",
            "IDX": 1,
            "z": [2, 3],
            "is_donor": False,
            "geometries": [
                {"phi": 0.0, "distance": 3.0},
            ],
        },
        {
            "smarts": "[#1][#7,#8,#9]",
            "IDX": 1,
            "z": [2],
            "is_donor": True,
            "geometries": [
                {"phi": 0.0, "distance": 2.0},
            ],
        },
        {
            "smarts": "[#8X1]=[X3][*]",
            "IDX": 1,
            "z": [2],
            "x": [3],
            "is_donor": False,
            "geometries": [
                {"phi": 60.0, "theta": 0, "distance": 3.0},
                {"phi": 60.0, "theta": 180, "distance": 3.0},
            ],
        },
    ]

    def __init__(self, water_model="tip3p", planar_tol=0.05):
        self.rule_list = json.loads(json.dumps(self.defaults))
        self.planar_tol = planar_tol
        if water_model == "tip3p":
            self.make_water = Waters.make_molsetup_TIP3P_AA
        else:
            raise RuntimeError("unknown water model: %s" % water_model)

    def __call__(self, molsetup: RDKitMoleculeSetup):
        coordinates = [molsetup.get_coord(i) for i in range(molsetup.true_atom_count)]
        water_molsetup_list = []
        for rule in self.rule_list:
            hits = molsetup.find_pattern(rule["smarts"])
            if len(hits) == 0:
                continue
            matched_idxs = set()
            smarts_idx = rule.get("IDX", 1) - 1  # default to first atom in SMARTS
            for hit in hits:
                parent_index = hit[smarts_idx]
                parent_center = coordinates[parent_index]
                if parent_index in matched_idxs:
                    warnings.warn(
                        "SMARTS <%s> matches same target atom more than once, ignoring"
                        % rule["smarts"]
                    )
                    continue
                matched_idxs.add(parent_index)
                # required settings
                z = [hit[i - 1] for i in rule["z"]]
                is_donor = rule["is_donor"]
                # optional settings
                x = rule.get("x", [])
                x = [hit[i - 1] for i in x]
                x90 = rule.get("x90", False)
                atomgeom = AtomicGeometry(parent_index, z, x, x90, self.planar_tol)
                for geometry in rule["geometries"]:
                    # required values
                    distance = geometry["distance"]
                    phi = np.radians(geometry["phi"])
                    # optional values
                    theta = geometry.get("theta", 0)
                    theta = np.radians(theta)
                    # place water
                    water_center = atomgeom.calc_point(
                        distance, theta, phi, coordinates
                    )
                    watersetup = self.make_water()
                    watercoords = watersetup.coord
                    self.orient_water(
                        watercoords, water_center, parent_center, is_donor
                    )
                    for i in range(len(watercoords)):
                        watersetup.coord[i] = watercoords[i]
                    water_molsetup_list.append(watersetup)
        return water_molsetup_list

    @staticmethod
    def orient_water(coords, target_xyz, anchor_xyz, anchor_is_donor):
        """
        Coordinates will be changed in place
        target_xyz is where the water oxygen will be
        anchor_xyz is the atom to which this molecule belongs
        anchor_is_donor is True for H, and False for O, N, S

        expects starting O to be at (0, 0, 0) and an H along x-axis

        Parameters
        ----------
        coords
        target_xyz
        anchor_xyz
        anchor_is_donor

        Returns
        -------

        """

        if anchor_is_donor:
            ang_hoh = np.radians(104.52)
            ang_lp = np.radians(109.5)  # probably good enough for lone pairs
            x = np.cos(np.pi - ang_hoh / 2) * np.cos(ang_lp / 2)
            y = -np.sin(np.pi - ang_hoh / 2) * np.cos(ang_lp / 2)
            z = np.sin(ang_lp / 2)
            axis = (x, y, z)
        else:
            axis = (1, 0, 0)
        # rotate
        v = np.array(anchor_xyz - target_xyz)
        v /= np.sqrt(np.dot(v, v))
        rotaxis = np.cross(axis, v)
        magnitude = np.sqrt(np.dot(rotaxis, rotaxis))
        if magnitude > 1e-6:
            # rotangle = np.arccos(np.dot(axis, v))
            rotangle = np.arcsin(magnitude)
            for i in range(len(coords)):
                coords_tuple = AtomicGeometry.rot3D(
                    coords[i], rotaxis, rotangle
                )  # normalizes rotaxis
                coords[i] = list(coords_tuple)

        # translate
        for i in range(len(coords)):
            for j in range(3):
                coords[i][j] = coords[i][j] + target_xyz[j]


class HydrateMoleculeLegacy:

    DEFAULT_HYDRATE_DISTANCE = 3.0
    DEFAULT_HYDRATE_CHARGE = 0.0
    DEFAULT_ATOM_TYPE = "W"
    DEFAULT_HB_LENGTH = 3.0

    def __init__(
        self,
        distance: float = DEFAULT_HYDRATE_DISTANCE,
        charge: float = DEFAULT_HYDRATE_CHARGE,
        atom_type: str = DEFAULT_ATOM_TYPE,
    ):
        """
        Initialize the legacy hydrate typer for AutoDock 4.2.x

        Parameters
        ----------
        distance: float
            Distance between water molecules and ligand heavy atoms.
        charge: float
            Partial charge of the water molecule. Not use for the hydrated docking.
        atom_type: str
            Atom type of the water molecule.
        """
        self._distance = distance
        self._charge = charge
        self._atom_type = atom_type
        self._bond_type = 1
        self._rotatable = False
        self._hb_config = {
            "HD": {1: (1, 1)},  # neigh: 1, wat: 1, sp1
            "OA": {
                1: (2, 2),  # neigh: 1, wat: 2, sp2
                2: (2, 3),  # neigh: 2, wat: 2, sp3
            },
            "SA": {
                1: (2, 2),  # neigh: 1, wat: 2, sp2
                2: (2, 3),  # neigh: 2, wat: 2, sp3
            },
            "NA": {
                1: (1, 1),  # neigh: 1, wat: 3, sp1
                2: (1, 2),  # neigh: 2, wat: 1, sp2
                3: (1, 3),  # neigh: 3, wat: 1, sp3
            },
        }

    def _place_sp1_one_water(
        self, anchor_xyz, neighbor_xyz, hb_length=DEFAULT_HB_LENGTH
    ):
        position = anchor_xyz + geomutils.vector(neighbor_xyz, anchor_xyz)
        position = geomutils.resize_vector(position, hb_length, anchor_xyz)
        positions = np.array([position])

        return positions

    def _place_sp2_one_water(
        self, anchor_xyz, neighbor1_xyz, neighbor2_xyz, hb_length=DEFAULT_HB_LENGTH
    ):
        position = geomutils.atom_to_move(anchor_xyz, [neighbor1_xyz, neighbor2_xyz])
        position = geomutils.resize_vector(position, hb_length, anchor_xyz)
        positions = np.array([position])

        return positions

    def _place_sp2_two_waters(
        self, anchor_xyz, neighbor1_xyz, neighbor2_xyz, hb_lengths, angles
    ):
        if len(hb_lengths) != 2:
            raise ValueError()
        if len(angles) != 2:
            raise ValueError()

        positions = []

        r = geomutils.rotation_axis(
            neighbor1_xyz, anchor_xyz, neighbor2_xyz, origin=anchor_xyz
        )
        p = neighbor1_xyz

        # We rotate p to get each vectors if necessary
        for hb_length, angle in zip(hb_lengths, angles):
            vector = p
            if angle != 0.0:
                position = geomutils.rotate_point(vector, anchor_xyz, r, angle)
            position = geomutils.resize_vector(position, hb_length, anchor_xyz)
            positions.append(position)

        positions = np.array(positions)

        return positions

    def _place_sp3_one_water(
        self, anchor_xyz, neighbor1_xyz, neighbor2_xyz, neighbor3_xyz, hb_length
    ):
        # We have to normalize bonds, otherwise the water molecule is not well placed
        v1 = anchor_xyz + geomutils.normalize(
            geomutils.vector(anchor_xyz, neighbor1_xyz)
        )
        v2 = anchor_xyz + geomutils.normalize(
            geomutils.vector(anchor_xyz, neighbor2_xyz)
        )
        v3 = anchor_xyz + geomutils.normalize(
            geomutils.vector(anchor_xyz, neighbor3_xyz)
        )

        position = geomutils.atom_to_move(anchor_xyz, [v1, v2, v3])
        position = geomutils.resize_vector(position, hb_length, anchor_xyz)
        positions = np.array([position])

        return positions

    def _place_sp3_two_waters(
        self, anchor_xyz, neighbor1_xyz, neighbor2_xyz, hb_lengths, angles
    ):
        if len(hb_lengths) != 2:
            raise ValueError()
        if len(angles) != 2:
            raise ValueError()

        positions = []

        v1 = anchor_xyz + geomutils.normalize(
            geomutils.vector(anchor_xyz, neighbor1_xyz)
        )
        v2 = anchor_xyz + geomutils.normalize(
            geomutils.vector(anchor_xyz, neighbor2_xyz)
        )

        r = anchor_xyz + geomutils.normalize(geomutils.vector(v1, v2))
        p = geomutils.atom_to_move(anchor_xyz, [v1, v2])

        # We rotate p to get each vectors if necessary
        for hb_length, angle in zip(hb_lengths, angles):
            vector = p
            if angle != 0.0:
                position = geomutils.rotate_point(vector, anchor_xyz, r, angle)
            position = geomutils.resize_vector(position, hb_length, anchor_xyz)
            positions.append(position)

        positions = np.array(positions)

        return positions

    def hydrate(self, setup: RDKitMoleculeSetup):
        """Add water molecules to the ligand

        Args:
            setup: MoleculeSetup object

        """
        water_anchors = []
        water_positions = []
        # It will be the same distance for all of the water molecules
        hb_length = self._distance

        for atom in setup.atoms:
            neighbors = atom.graph
            atom_type = atom.atom_type
            anchor_xyz = atom.coord
            neighbor1_xyz = setup.get_coord(neighbors[0])
            positions = np.array([])
            n_wat = None
            hyb = None

            if atom_type in self._hb_config:
                try:
                    n_wat, hyb = self._hb_config[atom_type][len(neighbors)]
                except KeyError:
                    raise RuntimeError(
                        "Cannot place water molecules on atom %d of type %s with %d neighbors."
                        % (atom.index, atom_type, len(neighbors))
                    )

                water_anchors.append(atom.index)

            if hyb == 1:
                if n_wat == 1:
                    # Example: X-HD
                    positions = self._place_sp1_one_water(
                        anchor_xyz, neighbor1_xyz, hb_length - 1.0
                    )
            elif hyb == 2:
                if n_wat == 1:
                    # Example: X-Nitrogen-X
                    neighbor2_xyz = setup.get_coord(neighbors[1])
                    positions = self._place_sp2_one_water(
                        anchor_xyz, neighbor1_xyz, neighbor2_xyz, hb_length
                    )
                elif n_wat == 2:
                    # Example: C=0 (backbone oxygen)
                    tmp_neighbors = [
                        x
                        for x in setup.get_neighbors(neighbors[0])
                        if not x == atom.index
                    ]
                    neighbor2_xyz = setup.get_coord(tmp_neighbors[0])
                    positions = self._place_sp2_two_waters(
                        anchor_xyz,
                        neighbor1_xyz,
                        neighbor2_xyz,
                        [hb_length, hb_length],
                        [-np.radians(120), np.radians(120)],
                    )
                elif n_wat == 3:
                    hyb = 3
            elif hyb == 3:
                if n_wat == 1:
                    # Example: Ammonia
                    neighbor2_xyz = setup.get_coord(neighbors[1])
                    neighbor3_xyz = setup.get_coord(neighbors[2])
                    positions = self._place_sp3_one_water(
                        anchor_xyz,
                        neighbor1_xyz,
                        neighbor2_xyz,
                        neighbor3_xyz,
                        hb_length,
                    )
                elif n_wat == 2:
                    # Example: O-HD (Oxygen in hydroxyl group)
                    neighbor2_xyz = setup.get_coord(neighbors[1])
                    positions = self._place_sp3_two_waters(
                        anchor_xyz,
                        neighbor1_xyz,
                        neighbor2_xyz,
                        [hb_length, hb_length],
                        [-np.radians(60), np.radians(60)],
                    )
                elif n_wat == 3:
                    positions = np.array([])

            if positions.size:
                water_positions.append(positions)

        for water_anchor, waters_on_anchor in zip(water_anchors, water_positions):
            for water_on_anchor in waters_on_anchor:
                tmp = setup.get_pdbinfo(water_anchor)
                pdbinfo = pdbutils.PDBAtomInfo(
                    "WAT", tmp.resName, tmp.resNum, tmp.icode, tmp.chain
                )
                setup.add_pseudoatom(
                    coord=water_on_anchor,
                    charge=self._charge,
                    anchor_list=[water_anchor],
                    atom_type=self._atom_type,
                    rotatable=self._rotatable,
                    pdbinfo=pdbinfo,
                )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import warnings

import numpy as np

from .utils import pdbutils


class AtomTyper:

    @classmethod
    def type_everything(
        cls,
        molsetup,
        atom_params,
        charge_model,
        offatom_params=None,
        dihedral_params=None,
    ):

        cls._type_atoms(molsetup, atom_params)

        # offatoms must be typed after charges, because offsites pull charge
        if offatom_params is not None:
            cached_offatoms = cls._cache_offatoms(molsetup, offatom_params)
            coords = {atom.index: atom.coord for atom in molsetup.atoms if not atom.is_dummy}
            cls._set_offatoms(molsetup, cached_offatoms, coords)

        if dihedral_params not in (None, "espaloma"):
            cls._type_dihedrals(molsetup, dihedral_params)

        return

    @staticmethod
    def _type_atoms(molsetup, atom_params):
        # ensure every "atompar" is defined in a single "smartsgroup"
        ensure = {}
        # go over all "smartsgroup"s
        for smartsgroup in atom_params:
            if smartsgroup == "comment":
                continue
            for line in atom_params[
                smartsgroup
            ]:  # line is a dict, e.g. {"smarts": "[#1][#7,#8,#9,#15,#16]","atype": "HD"}
                smarts = str(line["smarts"])
                # get indices of the atoms in the smarts to which the parameters will be assigned
                idxs = [
                    0
                ]  # by default, the first atom in the smarts gets parameterized
                if "IDX" in line:
                    idxs = [i for i in line["IDX"]]
                # match SMARTS
                hits = molsetup.find_pattern(smarts)
                for atompar in line:
                    if atompar in ["smarts", "comment", "IDX"]:
                        continue
                    if atompar not in molsetup.atom_params:
                        molsetup.atom_params[atompar] = [None] * len(molsetup.atoms)
                    value = line[atompar]
                    # keep track of every "smartsgroup" that modified "atompar"
                    ensure.setdefault(atompar, [])
                    ensure[atompar].append(smartsgroup)
                    # Each "hit" is a tuple of atom indices that matched the smarts
                    # The length of each "hit" is the number of atoms in the smarts
                    for hit in hits:
                        # Multiple atoms may be targeted by a single smarts:
                        # For example: both oxygens in NO2 are parameterized by a single smarts pattern.
                        # "idxs" are 1-indices of atoms in the smarts to which parameters are to be assigned.
                        for idx in idxs:
                            if atompar == "atype":
                                molsetup.set_atom_type(
                                    hit[idx], value
                                )  # overrides previous calls
                            molsetup.atom_params[atompar][hit[idx]] = value

        # guarantee that each atompar is exclusive of a single group
        for atompar in ensure:
            if len(set(ensure[atompar])) > 1:
                msg = "%s is modified in multiple smartsgroups: %s" % (
                    atompar,
                    set(ensure[atompar]),
                )
                warnings.warn(msg)
        return

    @staticmethod
    def _cache_offatoms(molsetup, offatom_params):
        """precalculate off-site atoms"""
        cached_offatoms = {}
        n_offatoms = 0
        atoms_with_offchrg = set()
        # each parent atom can only be matched once in each smartsgroup
        for smartsgroup in offatom_params:
            if smartsgroup == "comment":
                continue
            tmp = {}
            for line in offatom_params[smartsgroup]:
                # SMARTS
                smarts = str(line["smarts"])
                hits = molsetup.find_pattern(smarts)
                # atom indexes in smarts string
                smarts_idxs = [0]
                if "IDX" in line:
                    smarts_idxs = [i for i in line["IDX"]]
                for smarts_idx in smarts_idxs:
                    for hit in hits:
                        parent_idx = hit[smarts_idx]
                        tmp.setdefault(
                            parent_idx, []
                        )  # TODO tmp[parent_idx] = [], yeah?
                        for offatom in line["OFFATOMS"]:
                            # set defaults
                            tmp[parent_idx].append(
                                {
                                    "offatom": {
                                        "distance": 1.0,
                                        "x90": False,
                                        "phi": 0.0,
                                        "theta": 0.0,
                                        "z": [],
                                        "x": [],
                                    },
                                    "atom_params": {},
                                }
                            )
                            for key in offatom:
                                if key in ["distance", "x90"]:
                                    tmp[parent_idx][-1]["offatom"][key] = offatom[key]
                                # replace SMARTS indexes by the atomic index
                                elif key in ["z", "x"]:
                                    for i in offatom[key]:
                                        idx = hit[i]
                                        tmp[parent_idx][-1]["offatom"][key].append(idx)
                                # convert degrees to radians
                                elif key in ["theta", "phi"]:
                                    tmp[parent_idx][-1]["offatom"][key] = np.radians(
                                        offatom[key]
                                    )
                                # ignore comments
                                elif key in ["comment"]:
                                    pass
                                elif key == "atype":
                                    tmp[parent_idx][-1]["atom_params"][key] = offatom[
                                        key
                                    ]
                                elif key == "pull_charge_fraction":
                                    if parent_idx in atoms_with_offchrg:
                                        raise RuntimeError(
                                            "atom %d has charge pulled more than once"
                                            % parent_idx
                                        )
                                    atoms_with_offchrg.add(parent_idx)
                                    tmp[parent_idx][-1]["atom_params"][key] = offatom[
                                        key
                                    ]
                                else:
                                    pass
            for parent_idx in tmp:
                for offatom_dict in tmp[parent_idx]:
                    atom_params = offatom_dict["atom_params"]
                    offatom = offatom_dict["offatom"]
                    atomgeom = AtomicGeometry(
                        parent_idx,
                        neigh=offatom["z"],
                        xneigh=offatom["x"],
                        x90=offatom["x90"],
                    )
                    if "pull_charge_fraction" in atom_params:
                        pull_charge_fraction = atom_params["pull_charge_fraction"]
                    else:
                        pull_charge_fraction = 0.0
                    args = (
                        atom_params["atype"],
                        offatom["distance"],
                        offatom["theta"],
                        offatom["phi"],
                        pull_charge_fraction,
                    )
                    # number of coordinates (before adding new offatom)
                    cached_offatoms[n_offatoms] = (atomgeom, args)
                    n_offatoms += 1
        return cached_offatoms

    @staticmethod
    def _set_offatoms(molsetup, cached_offatoms, coords):
        """add cached offatoms"""
        for k, (atomgeom, args) in cached_offatoms.items():
            (atom_type, dist, theta, phi, pull_charge_fraction) = args
            offatom_coords = atomgeom.calc_point(dist, theta, phi, coords)
            tmp = molsetup.get_pdbinfo(atomgeom.parent + 1)
            pdbinfo = pdbutils.PDBAtomInfo(
                "G", tmp.resName, tmp.resNum, tmp.icode, tmp.chain
            )
            q_parent = (1 - pull_charge_fraction) * molsetup.get_charge(atomgeom.parent)
            q_offsite = pull_charge_fraction * molsetup.get_charge(atomgeom.parent)
            pseudo_atom = {
                "coord": offatom_coords,
                "anchor_list": [atomgeom.parent],
                "charge": q_offsite,
                "pdbinfo": pdbinfo,
                "atom_type": atom_type,
                "rotatable": False,
            }
            molsetup.atoms[atomgeom.parent].charge = q_parent
            molsetup.add_pseudoatom(**pseudo_atom)
        return

    @staticmethod
    def _type_dihedrals(molsetup, dihedral_params):

        dihedrals = {}

        for line in dihedral_params:
            smarts = str(line["smarts"])
            hits = molsetup.find_pattern(smarts)
            if len(hits) == 0:
                continue  # non-rotatable bonds get dihedrals
            idxs = [i for i in line["IDX"]]
            tid = line["id"] if "id" in line else None
            fourier_series = []
            term_indices = {}
            for key in line:
                for keyword in ["phase", "k", "periodicity", "idivf"]:
                    if key.startswith(keyword):
                        t = int(key.replace(keyword, ""))  # e.g. "phase2" -> int(2)
                        if t not in term_indices:
                            term_indices[t] = len(fourier_series)
                            fourier_series.append({})
                        index = term_indices[t]
                        fourier_series[index][keyword] = line[key]
                        break

            for index in range(len(fourier_series)):
                if "idivf" in fourier_series[index]:
                    idivf = fourier_series[index].pop("idivf")
                    fourier_series[index]["k"] /= idivf

            dihedral_index = molsetup.add_dihedral_interaction(fourier_series)

            for hit in hits:
                atom_idxs = tuple([hit[j] for j in idxs])
                molsetup.dihedral_partaking_atoms[atom_idxs] = dihedral_index
                molsetup.dihedral_labels[atom_idxs] = tid


class AtomicGeometry:
    """generate reference frames and add extra sites"""

    def __init__(self, parent, neigh, xneigh=[], x90=False, planar_tol=0.1):
        """arguments are indices of atoms"""

        self.planar_tol = planar_tol  # angstroms, length of neighbor vecs for z-axis

        # real atom hosting extra sites
        if type(parent) != int:
            raise RuntimeError("parent must be int")
        self.parent = parent

        # list of bonded atoms (used to define z-axis)
        self.neigh = []
        for i in neigh:
            if type(i) != int:
                raise RuntimeError("neigh indices must be int")
            self.neigh.append(i)

        # list of atoms that
        self.xneigh = []
        for i in xneigh:
            if type(i) != int:
                raise RuntimeError("xneigh indices must be int")
            self.xneigh.append(i)

        self.calc_x = len(self.xneigh) > 0
        self.x90 = x90  # y axis becomes x axis (useful to rotate in-plane by 90 deg)

    def calc_point(self, distance, theta, phi, coords):
        """return coordinates of point specified in spherical coordinates"""

        z = self._calc_z(coords)

        # return pt aligned with z-axis
        if phi == 0.0:
            return z * distance + np.array(coords[self.parent])

        # need x-vec if phi != 0
        elif self.calc_x == False:
            raise RuntimeError("phi must be zero if X undefined")

        else:
            x = self._calc_x(coords)
            if self.x90:
                x = np.cross(self.z, x)
            y = np.cross(z, x)
            pt = z * distance
            pt = self.rot3D(pt, y, phi)
            pt = self.rot3D(pt, z, theta)
            pt += np.array(coords[self.parent])
            return pt

    def _calc_z(self, coords):
        """maximize distance from neigh"""
        z = np.zeros(3)
        cumsum = np.zeros(3)
        for i in self.neigh:
            v = np.array(coords[self.parent]) - np.array(coords[i])
            cumsum += v
            z += self.normalized(v)
        z = self.normalized(z)
        if np.sum(cumsum**2) < self.planar_tol**2:
            raise RuntimeError("Refusing to place Z axis on planar atom")
        return z

    def _calc_x(self, coords):
        x = np.zeros(3)
        for i in self.xneigh:
            v = np.array(coords[self.parent]) - np.array(coords[i])
            x += self.normalized(v)
        x = self.normalized(x)
        return x

    @staticmethod
    def rot3D(pt, ax, rad):
        """
        Rotate point:
        pt = (x,y,z) coordinates to be rotated
        ax = vector around wich rotation is performed
        rad = rotate by "rad" radians
        """
        # If axis has len=0, rotate by 0.0 rad on any axis
        # Make sure ax has unitary length
        len_ax = (ax[0] ** 2 + ax[1] ** 2 + ax[2] ** 2) ** 0.5
        if len_ax == 0.0:
            u, v, w = (1, 0, 0)
            rad = 0.0
        else:
            u, v, w = [i / len_ax for i in ax]
        x, y, z = pt
        ux, uy, uz = u * x, u * y, u * z
        vx, vy, vz = v * x, v * y, v * z
        wx, wy, wz = w * x, w * y, w * z
        sa = np.sin(rad)
        ca = np.cos(rad)
        p0 = (
            u * (ux + vy + wz)
            + (x * (v * v + w * w) - u * (vy + wz)) * ca
            + (-wy + vz) * sa
        )
        p1 = (
            v * (ux + vy + wz)
            + (y * (u * u + w * w) - v * (ux + wz)) * ca
            + (wx - uz) * sa
        )
        p2 = (
            w * (ux + vy + wz)
            + (z * (u * u + v * v) - w * (ux + vy)) * ca
            + (-vx + uy) * sa
        )
        return (p0, p1, p2)

    def normalized(self, vec):
        l = sum([x**2 for x in vec]) ** 0.5
        if type(vec) == list:
            return [x / l for x in vec]
        else:
            # should be np.array
            return vec / l



from meeko import MoleculeSetup

import pathlib
import pytest

import meeko


# region Fixtures
# We may want to find a different place for these as tests get more complex structure and revisions.
@pytest.fixture
def empty_molecule_setup():
    molsetup = MoleculeSetup()
    yield molsetup


# endregion


class TestMoleculeSetupInit:
    def test_add_one_atom(self):
        pass

    def test_add_one_pseudo_atom(self):
        pass

    def test_add_multiple_atoms(self):
        pass

    def test_add_bonds(self):
        pass



#!/usr/bin/env python

import pathlib
from rdkit import Chem
from meeko import MoleculePreparation


workdir = pathlib.Path(__file__).parents[0]


def run(
    folder,
    fn,
    expected_rot_bonds,
    expected_sizes_of_rigid_bodies,
    expected_members=None,
):

    path = workdir / pathlib.Path(folder) / fn
    mol = Chem.MolFromMolFile(str(path), removeHs=False)
    mk_prep = MoleculePreparation()
    molsetup = mk_prep(mol)[0]

    # sizes of rigid groups/bodies
    rigid_body_members = molsetup.flexibility_model["rigid_body_members"]
    sizes = []
    for _, members in rigid_body_members.items():
        sizes.append(len(members))
    assert sorted(sizes) == sorted(expected_sizes_of_rigid_bodies)

    # check rotatable bonds
    rot_bonds = set()
    for bond_id, bond in molsetup.bond_info.items():
        if bond.rotatable:
            rot_bonds.add(bond_id)
    assert rot_bonds == expected_rot_bonds

    # optionally check actual atom indices in each rigid body
    if expected_members is not None:
        expected_members = set([tuple(sorted(members)) for members in expected_members])
        actual_members = set()
        for _, members in rigid_body_members.items():
            actual_members.add(tuple(sorted(members)))
        assert actual_members == expected_members
    return


def test_non_sequential_atom_ordering_01():
    run(
        fn="non_sequential_atom_ordering_01.mol",
        folder="flexibility_data",
        expected_rot_bonds={(0, 1)},
        expected_sizes_of_rigid_bodies=[2, 8],
    )


def test_non_sequential_atom_ordering_02():
    run(
        fn="non_sequential_atom_ordering_02.mol",
        folder="flexibility_data",
        expected_rot_bonds={(0, 1), (1, 2), (4, 6), (6, 7)},
        expected_sizes_of_rigid_bodies=[1, 1, 2, 6, 7],
        expected_members=[
            [1],
            [6],
            [2, 3],
            [7, 8, 9, 10, 11, 12],
            [0, 4, 5, 13, 14, 15, 16],
        ],
    )


def test_non_sequential_atom_ordering_03():
    run(
        fn="non_sequential_atom_ordering_03.mol",
        folder="macrocycle_data",
        expected_rot_bonds={(0, 1), (1, 2), (2, 3), (3, 8), (4, 6), (6, 7)},
        expected_sizes_of_rigid_bodies=[1, 1, 2, 2, 6, 7],
        expected_members=[
            [1],
            [2, 32],
            [3, 33],
            [6],
            [0, 4, 5, 16, 14, 13, 15],
            [7, 8, 9, 10, 11, 12],
        ],
    )


def test_non_sequential_atom_ordering_04():
    """this one lacks aliphatic C-C bond and can't break under CG/G typing"""
    run(
        fn="non_sequential_atom_ordering_04.mol",
        folder="macrocycle_data",
        expected_rot_bonds=set(),
        expected_sizes_of_rigid_bodies=[19],
    )


def test_non_sequential_atom_ordering_05():
    """two rings, one breakable, and one lacking an aliphatic C-C bond
    that can't break under CG/G typing"""
    run(
        fn="non_sequential_atom_ordering_05.mol",
        folder="macrocycle_data",
        expected_rot_bonds={(14, 16), (16, 17), (17, 22), (10, 11), (11, 19)},
        expected_sizes_of_rigid_bodies=[17, 1, 5, 2, 2],
    )
